from ..ds9_attr_parser import get_ds9_attr_parser, get_attr, Ds9AttrParser


def test_attr():
    p = get_ds9_attr_parser()
    assert p.parseString("color=green")[0] == ("color", "green")
    assert p.parseString("font=\"123 123\"")[0] == ("font", '"123 123"')
    assert p.parseString("color")[0] == ("color",)
    assert p.parseString("tag={group 1}")[0] == ("tag", "group 1")
    assert p.parseString('color=#6a8')[0] == ("color", "#6a8")


def test_get_attr():
    attr_list = [('tag', 'group1'), ('tag', 'group2'),
                 ('tag', 'group3'), ('color', 'green')]
    global_attrs = [], {}

    attr = get_attr(attr_list, global_attrs)
    assert attr[0] == []
    assert attr[1] == {'color': 'green',
                       'tag': set(['group1', 'group3', 'group2'])}


def test_shape_in_comment():
    parser = Ds9AttrParser()

    r = parser.parse_check_shape("segment(0, 2)")
    assert r[0].name == "segment"
    assert r[1] == []

    r = parser.parse_check_shape("projection(0, 2, 3, 2, 4)")
    assert r[0].name == "projection"
    assert r[1] == []



from ..parser_helper import define_shape


def test_define_shape():
    from ..region_numbers import CoordOdd, CoordEven, Distance
    args = [s.parser for s in [CoordOdd, CoordEven, Distance]]
    circle_parser = define_shape("circle", args, args_repeat=None)

    p = circle_parser.parseString("circle(1:2:3, 2:4:5, 3.)")
    assert p[0] == "circle"
    assert isinstance(p[1], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)

    args = [s.parser for s in [CoordEven]]
    circle_parser = define_shape("circle", args, args_repeat=None)

    p = circle_parser.parseString("circle(1:2:3)")
    p = circle_parser.parseString("circle(1d2m3s)")
    assert p[0] == "circle"
    assert isinstance(p[1], CoordEven.type)

    args = [s.parser for s in [CoordOdd, CoordEven, Distance, Distance]]
    ell_parser = define_shape("ell", args, args_repeat=(2, 4))

    p = ell_parser.parseString("ell(1:2:2, 2:2:2, 3, 4, 5, 6)")
    assert p[0] == "ell"
    assert isinstance(p[1], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)

    args = [s.parser for s in [CoordOdd, CoordEven]]
    polygon_parser = define_shape("polygon", args, args_repeat=(0, 2))
    p = polygon_parser.parseString("polygon(3:2:4.22, 3:3:4., 3:2:3, 3.2)")
    assert p[0] == "polygon"
    assert isinstance(p[-2], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)



"""
pyregion: a Python parser for ds9 region files

* Code : https://github.com/astropy/pyregion
* Docs : http://pyregion.readthedocs.io/

See also the in-development ``regions`` package
at https://github.com/astropy/regions
a new astronomy package for regions based on Astropy.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    from .core import *
    from .parser_helper import Shape



from ..ds9_region_parser import RegionParser, Global
from ..parser_helper import CoordCommand, Shape
from ..region_numbers import SimpleNumber, AngularDistance


def test_regionLine():
    test_string_1 = [
        "circle(109,253,28.950304) # comment 1",
        "polygon(257,290,300.78944,271.78944,300.78944,178.21056,258,216,207.21056,178.21056)",
        "polygon(273.98971,175.01029,274.01029,175.01029,274.01029,174.98971,273.98971,174.98971)",
        "-rotbox(162,96.5,134,41,43.801537)",
        "ellipse(172,328,23,41,27.300813)",
    ]

    test_names = [
        "circle",
        "polygon",
        "polygon",
        "rotbox",
        "ellipse",
    ]

    rp = RegionParser()

    for s, n in zip(test_string_1, test_names):
        s = rp.parseLine(s)[0]
        assert len(s) == 1
        assert s[0].name == n


def test_comment():
    s = "circle(3323, 423, 423) # comment"

    rp = RegionParser()
    c = rp.parseLine(s)[1]

    assert c == "comment"

    s = " # comment2"
    c = rp.parseLine(s)[1]

    assert c == "comment2"


def test_global():
    s = 'global color=green font="helvetica 10 normal" select=1 highlite=1 edit=1 move=1 delete=1 include=1 fixed=0 source'

    rp = RegionParser()
    ss = rp.parseLine(s)[0]

    assert isinstance(ss[0], Global)


def test_space_delimited_region():
    """
    Regression test for https://github.com/astropy/pyregion/issues/73
    """
    s = 'J2000; circle 188.5557102 12.0314056 1" # color=red'

    rp = RegionParser()
    ss = rp.parseLine(s)[0]

    assert isinstance(ss[0], CoordCommand)
    assert ss[0].text == "J2000"

    assert isinstance(ss[1], Shape)
    param_types = list(map(type, ss[1].params))
    assert param_types == [SimpleNumber, SimpleNumber, AngularDistance]



import os
from os.path import join
from astropy.io.fits import Header
from .. import parse
from numpy.testing import assert_allclose

rootdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


def demo_header():
    return Header.fromtextfile(join(rootdir, "sample_fits04.header"))


def test_cube():
    header = demo_header()

    region_string = 'circle(12:04:15.065,+18:26:51.00,173.029")'
    r = parse(region_string).as_imagecoord(header)

    assert_allclose(r[0].coord_list, [117, 132, 34.6], atol=0.01)



from itertools import cycle

from .ds9_region_parser import RegionParser
from .wcs_converter import check_wcs as _check_wcs

_builtin_open = open


class ShapeList(list):
    """A list of `~pyregion.Shape` objects.

    Parameters
    ----------
    shape_list : list
        List of `pyregion.Shape` objects
    comment_list : list, None
        List of comment strings for each argument
    """

    def __init__(self, shape_list, comment_list=None):
        if comment_list is not None:
            if len(comment_list) != len(shape_list):
                err = "Ambiguous number of comments {} for number of shapes {}"
                raise ValueError(err.format(len(comment_list),
                                            len(shape_list)))
        self._comment_list = comment_list
        list.__init__(self, shape_list)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return ShapeList(list.__getitem__(self, key))
        else:
            return list.__getitem__(self, key)

    def __getslice__(self, i, j):
        return self[max(0, i):max(0, j):]

    def check_imagecoord(self):
        """Are all shapes in image coordinates?

        Returns ``True`` if yes, and ``False`` if not.
        """
        if [s for s in self if s.coord_format != "image"]:
            return False
        else:
            return True

    def as_imagecoord(self, header):
        """New shape list in image coordinates.

        Parameters
        ----------
        header : `~astropy.io.fits.Header`
            FITS header

        Returns
        -------
        shape_list : `ShapeList`
            New shape list, with coordinates of the each shape
            converted to the image coordinate using the given header
            information.
        """

        comment_list = self._comment_list
        if comment_list is None:
            comment_list = cycle([None])

        r = RegionParser.sky_to_image(zip(self, comment_list),
                                      header)
        shape_list, comment_list = zip(*list(r))
        return ShapeList(shape_list, comment_list=comment_list)

    def get_mpl_patches_texts(self, properties_func=None,
                              text_offset=5.0,
                              origin=1):
        """
        Often, the regions files implicitly assume the lower-left
        corner of the image as a coordinate (1,1). However, the python
        convetion is that the array index starts from 0. By default
        (``origin=1``), coordinates of the returned mpl artists have
        coordinate shifted by (1, 1). If you do not want this shift,
        use ``origin=0``.
        """
        from .mpl_helper import as_mpl_artists
        patches, txts = as_mpl_artists(self, properties_func,
                                       text_offset,
                                       origin=origin)

        return patches, txts

    def get_filter(self, header=None, origin=1):
        """Get filter.
        Often, the regions files implicitly assume the lower-left
        corner of the image as a coordinate (1,1). However, the python
        convetion is that the array index starts from 0. By default
        (``origin=1``), coordinates of the returned mpl artists have
        coordinate shifted by (1, 1). If you do not want this shift,
        use ``origin=0``.

        Parameters
        ----------
        header : `astropy.io.fits.Header`
            FITS header
        origin : {0, 1}
            Pixel coordinate origin

        Returns
        -------
        filter : TODO
            Filter object
        """

        from .region_to_filter import as_region_filter

        if header is None:
            if not self.check_imagecoord():
                raise RuntimeError("the region has non-image coordinate. header is required.")
            reg_in_imagecoord = self
        else:
            reg_in_imagecoord = self.as_imagecoord(header)

        region_filter = as_region_filter(reg_in_imagecoord, origin=origin)

        return region_filter

    def get_mask(self, hdu=None, header=None, shape=None):
        """Create a 2-d mask.

        Parameters
        ----------
        hdu : `astropy.io.fits.ImageHDU`
            FITS image HDU
        header : `~astropy.io.fits.Header`
            FITS header
        shape : tuple
            Image shape

        Returns
        -------
        mask : `numpy.array`
            Boolean mask

        Examples
        --------
        get_mask(hdu=f[0])
        get_mask(shape=(10,10))
        get_mask(header=f[0].header, shape=(10,10))
        """

        if hdu and header is None:
            header = hdu.header
        if hdu and shape is None:
            shape = hdu.data.shape

        region_filter = self.get_filter(header=header)
        mask = region_filter.mask(shape)

        return mask

    def write(self, outfile):
        """Write this shape list to a region file.

        Parameters
        ----------
        outfile : str
            File name
        """
        if len(self) < 1:
            print("WARNING: The region list is empty. The region file "
                  "'{:s}' will be empty.".format(outfile))
            try:
                outf = _builtin_open(outfile, 'w')
                outf.close()
                return
            except IOError as e:
                cmsg = "Unable to create region file '{:s}'.".format(outfile)
                if e.args:
                    e.args = (e.args[0] + '\n' + cmsg,) + e.args[1:]
                else:
                    e.args = (cmsg,)
                raise e

        prev_cs = self[0].coord_format

        outf = None
        try:
            outf = _builtin_open(outfile, 'w')

            attr0 = self[0].attr[1]
            defaultline = " ".join(["{:s}={:s}".format(a, attr0[a])
                                    for a in attr0 if a != 'text'])

            # first line is globals
            outf.write("global {0}\n".format(defaultline))
            # second line must be a coordinate format
            outf.write("{0}\n".format(prev_cs))

            for shape in self:
                shape_attr = '' if prev_cs == shape.coord_format \
                    else shape.coord_format + "; "
                shape_excl = '-' if shape.exclude else ''
                text_coordlist = ["{:f}".format(f) for f in shape.coord_list]
                shape_coords = "(" + ",".join(text_coordlist) + ")"
                shape_comment = " # " + shape.comment if shape.comment else ''

                shape_str = (shape_attr + shape_excl + shape.name +
                             shape_coords + shape_comment)

                outf.write("{0}\n".format(shape_str))

        except IOError as e:
            cmsg = "Unable to create region file \'{:s}\'.".format(outfile)
            if e.args:
                e.args = (e.args[0] + '\n' + cmsg,) + e.args[1:]
            else:
                e.args = (cmsg,)
            raise e
        finally:
            if outf:
                outf.close()


def parse(region_string):
    """Parse DS9 region string into a ShapeList.

    Parameters
    ----------
    region_string : str
        Region string

    Returns
    -------
    shapes : `ShapeList`
        List of `~pyregion.Shape`
    """
    rp = RegionParser()
    ss = rp.parse(region_string)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)

    shape_list, comment_list = rp.filter_shape2(sss2)
    return ShapeList(shape_list, comment_list=comment_list)


def open(fname):
    """Open, read and parse DS9 region file.

    Parameters
    ----------
    fname : str
        Filename

    Returns
    -------
    shapes : `ShapeList`
        List of `~pyregion.Shape`
    """
    with _builtin_open(fname) as fh:
        region_string = fh.read()
    return parse(region_string)


def read_region(s):
    """Read region.

    Parameters
    ----------
    s : str
        Region string

    Returns
    -------
    shapes : `ShapeList`
        List of `~pyregion.Shape`
    """
    rp = RegionParser()
    ss = rp.parse(s)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)

    shape_list = rp.filter_shape(sss2)
    return ShapeList(shape_list)


def read_region_as_imagecoord(s, header):
    """Read region as image coordinates.

    Parameters
    ----------
    s : str
        Region string
    header : `~astropy.io.fits.Header`
        FITS header

    Returns
    -------
    shapes : `~pyregion.ShapeList`
        List of `~pyregion.Shape`
    """
    rp = RegionParser()
    ss = rp.parse(s)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)
    sss3 = rp.sky_to_image(sss2, header)

    shape_list = rp.filter_shape(sss3)
    return ShapeList(shape_list)


def get_mask(region, hdu, origin=1):
    """Get mask.

    Parameters
    ----------
    region : `~pyregion.ShapeList`
        List of `~pyregion.Shape`
    hdu : `~astropy.io.fits.ImageHDU`
        FITS image HDU
    origin : float
        TODO: document me

    Returns
    -------
    mask : `~numpy.array`
        Boolean mask

    Examples
    --------
    >>> from astropy.io import fits
    >>> from pyregion import read_region_as_imagecoord, get_mask
    >>> hdu = fits.open("test.fits")[0]
    >>> region = "test01.reg"
    >>> reg = read_region_as_imagecoord(open(region), f[0].header)
    >>> mask = get_mask(reg, hdu)
    """
    from pyregion.region_to_filter import as_region_filter

    data = hdu.data
    region_filter = as_region_filter(region, origin=origin)
    mask = region_filter.mask(data)
    return mask
